import processing.serial.*;

String strImage="{EYNOIM}FFFFFFFFFFFFFFFFFFFFFFFF00000000000000000000000000000000000001FC000006A700001A0000003C0000002000000050000000D0000FBFFFFF0E6B72750000FC000000FC000000FC000000FC000000FC000001FC000001FE000001FC000001FC000001FE000001FE000001FE000003FE000001FF000001FF000003FF0080";
String IMAGE_START_TAG = "{EYNOIM}";
char IMAGE_END_TAG = 0x0A;

final int ROW_COUNT = 32;
final int PIXELS_PER_ROW = 32;

boolean [][]_image = new boolean[ROW_COUNT][PIXELS_PER_ROW];
Serial myPort = null;

int prevMillis = 0;
float fRate = 0;

final int PIXEL_SIZE = 15;
final int MARGINE = 10;
final int BORDER_WIDTH = 4; 

final int LEGEND_WIDTH = 500;
final int LEGEND_FONT_HEIGHT = 24;

final int MAS_CENTER_CROSS_LINE_WIDTH = 4;
final int MAS_CENTER_CROSS_WING = 50;

private class RegionOfInterest{
  public int left;
  public int right;
  public int bottom;
  public int top;
  RegionOfInterest(int _left, int _right, int _bottom, int _top){
    left = _left;
    right = _right;
    bottom = _bottom;
    top = _top;
  }
}

RegionOfInterest roi = new RegionOfInterest(1,PIXELS_PER_ROW-2,1,ROW_COUNT-2);

private class MassCenter{
  public float x;
  public float y;
  public float mass;
  MassCenter(float _x, float _y, float _mass){
    x = _x;
    y = _y;
    mass = _mass;
  }
}

MassCenter massCenter = new MassCenter(10.5,10.5,5);

void setup(){
  size(PIXELS_PER_ROW*PIXEL_SIZE+BORDER_WIDTH+MARGINE*2, ROW_COUNT*PIXEL_SIZE+BORDER_WIDTH+2*MARGINE*3+LEGEND_FONT_HEIGHT);
  printArray(Serial.list());
  if(Serial.list().length>0){
    String portName = Serial.list()[0];
    myPort = new Serial(this, portName, 115200);
    myPort.bufferUntil(IMAGE_END_TAG);
  }
}

void serialEvent(Serial p) {
  String _strBuff = p.readString();
  int _imgStrIndex = _strBuff.indexOf(IMAGE_START_TAG);
  if(_imgStrIndex>=0){
    strImage = _strBuff.substring(_imgStrIndex);
//    println(_imgStrIndex);
    fRate = 1000/(millis()-prevMillis);
    prevMillis = millis();
    redraw();
  }
}

void decodeImage(){
  int _bitIndex=0;
  int _rowIndex=0;
  int _pixelIndex=0;
  String _buff;
  int pixelCount = (strImage.length()-IMAGE_START_TAG.length())/2;
//  println(strImage.length()); 
//  println(pixelCount);
  for(int i = 0;i<pixelCount;i++){
    _buff=binary(unhex(strImage.substring(IMAGE_START_TAG.length()+i*2,IMAGE_START_TAG.length()+i*2+2)),8);
    for(_bitIndex=0;_bitIndex<8;_bitIndex++){
      if(_buff.charAt(_bitIndex)=='1') {
        _image[_rowIndex][_pixelIndex]=true;
      }
      else {
        _image[_rowIndex][_pixelIndex]=false;
      }
      _pixelIndex++;
//      println(_pixelIndex);
    }
    if(_pixelIndex>=PIXELS_PER_ROW-1){
      _pixelIndex=0;
      _rowIndex++;
    }
    if(_rowIndex>=ROW_COUNT-1){
//      println(strImage.substring(IMAGE_START_TAG.length()+i*2));
      break;
    }
//    println(_buff);
  }
}

void calcMassCenter(){
  float mcx = 0;
  float mcy = 0;
  int _count = 0;
  for(int _row=0;_row<ROW_COUNT;_row++){
    for(int _pixel=0;_pixel<PIXELS_PER_ROW;_pixel++){
      if(_image[_row][_pixel]){
        if(((_row<=(ROW_COUNT-roi.bottom-1)) && (_row>=(ROW_COUNT-roi.top-1))) && ((_pixel>=roi.left) && (_pixel<=roi.right))) {
          mcx = mcx + _pixel;
          mcy = mcy + _row;
          _count++;
        }
      }
    }
  }
  if(_count>0){
    mcx = mcx/_count;
    mcy = mcy/_count;
  }
  massCenter.x = mcx;
  massCenter.y = mcy;
  massCenter.mass = _count;
}

void draw(){
  background(255);
  fill(127);
  stroke(255,0,0);
  strokeWeight(4);
  rect(MARGINE,MARGINE,PIXELS_PER_ROW*PIXEL_SIZE+BORDER_WIDTH,ROW_COUNT*PIXEL_SIZE+BORDER_WIDTH);
  fill(255);
  stroke(0,0,255);
  strokeWeight(1);
  rect(MARGINE+(roi.left)*PIXEL_SIZE,MARGINE+(ROW_COUNT-roi.top-1)*PIXEL_SIZE,(roi.right-roi.left+1)*PIXEL_SIZE+BORDER_WIDTH,(roi.top-roi.bottom+1)*PIXEL_SIZE+BORDER_WIDTH);
  stroke(0,0,0);
  strokeWeight(1);
  fill(0);
  decodeImage();
  calcMassCenter();
  for(int _row=0;_row<ROW_COUNT;_row++){
    for(int _pixel=0;_pixel<PIXELS_PER_ROW;_pixel++){
      if(_image[_row][_pixel]){
        if(((_row<=(ROW_COUNT-roi.bottom-1)) && (_row>=(ROW_COUNT-roi.top-1))) && ((_pixel>=roi.left) && (_pixel<=roi.right))) {
          stroke(0);
          fill(0);
        }
        else {
          stroke(80);
          fill(80);
        }
        rect(MARGINE+BORDER_WIDTH/2-1+_pixel*PIXEL_SIZE,MARGINE+BORDER_WIDTH/2+_row*PIXEL_SIZE,PIXEL_SIZE,PIXEL_SIZE);
      }
    }
  }
  if(massCenter.mass>0.5) {
    strokeWeight(MAS_CENTER_CROSS_LINE_WIDTH);
    stroke(255,50,50);
    line(MARGINE+BORDER_WIDTH/2-1+(massCenter.x+0.5)*PIXEL_SIZE-MAS_CENTER_CROSS_WING,MARGINE+BORDER_WIDTH/2+(massCenter.y+0.5)*PIXEL_SIZE,MARGINE+BORDER_WIDTH/2-1+(massCenter.x+0.5)*PIXEL_SIZE+MAS_CENTER_CROSS_WING,MARGINE+BORDER_WIDTH/2+(massCenter.y+0.5)*PIXEL_SIZE);
    line(MARGINE+BORDER_WIDTH/2-1+(massCenter.x+0.5)*PIXEL_SIZE,MARGINE+BORDER_WIDTH/2+(massCenter.y+0.5)*PIXEL_SIZE-MAS_CENTER_CROSS_WING,MARGINE+BORDER_WIDTH/2-1+(massCenter.x+0.5)*PIXEL_SIZE,MARGINE+BORDER_WIDTH/2+(massCenter.y+0.5)*PIXEL_SIZE+MAS_CENTER_CROSS_WING);
  }
  strokeWeight(1);
  stroke(0);
  fill(0);
  text((new String("ROI:   left: "))+str(roi.left)+"   right:\t"+str(roi.right)+"   bottom:\t"+str(roi.bottom)+"   top:\t"+str(roi.top),MARGINE,ROW_COUNT*PIXEL_SIZE+BORDER_WIDTH+2*MARGINE,LEGEND_WIDTH,LEGEND_FONT_HEIGHT);
  text((new String("Mass center:   x: "))+str(massCenter.x)+"   y:\t"+str(massCenter.y)+"   mass:\t"+str(massCenter.mass),MARGINE,ROW_COUNT*PIXEL_SIZE+BORDER_WIDTH+3*MARGINE+LEGEND_FONT_HEIGHT,LEGEND_WIDTH,LEGEND_FONT_HEIGHT);
}

boolean shiftPressed = false;

void keyPressed() {
  if(key == CODED) {
    if(keyCode == SHIFT) {
      shiftPressed = true;
    }
    if(keyCode == RIGHT) {
      if(shiftPressed) {
        if(roi.right<PIXELS_PER_ROW-1){
          roi.right++;
        }
      }
      else {
        if(roi.left<roi.right) {
          roi.left++;
        }
      }
    }
    if(keyCode == LEFT) {
      if(shiftPressed) {
        if(roi.right>roi.left){
          roi.right--;
        }
      }
      else {
        if(roi.left>0) {
          roi.left--;
        }
      }
    }
    if(keyCode == UP) {
      if(shiftPressed) {
        if(roi.bottom<roi.top) {
          roi.bottom++;
        }
      }
      else {
        if(roi.top<ROW_COUNT-1) {
          roi.top++;
        }
      }
    }
    if(keyCode == DOWN) {
      if(shiftPressed) {
        if(roi.bottom>0) {
          roi.bottom--;
        }
      }
      else {
        if(roi.top>roi.bottom) {
          roi.top--;
        }
      }
    }
  }
}

void keyReleased() {
  if(key == CODED) {
    if(keyCode == SHIFT) {
      shiftPressed = false;
    }
  }
}
